#ifndef GL_TEXT_SGG_H
#define GL_TEXT_SGG_H


#include <simple_regexp.h>
#include <gl_sgg.h>

namespace sgg {

using namespace std_ex;

//so there is one utility that I think properly belongs in nehe, but that rather has to 
//depend on things in sgg.  After some thought, I have decided that the "right" thing
//to do is to put it in gl_sgg, though the decision does bother me...
Drawable_Rects text_to_rect(nehe::TextFormat info, const char *fmt, ...);

struct DrawableString : public string, public Drawable {
	nehe::TextFormat format;
	void draw() {
		nehe::glPrintf(format, c_str());
	}
	DrawableString(string s,nehe::TextFormat format) 
		: string(s), format(format) {}
};

struct null_pick {
	void pick() {}	
};

//can't inherit from DrawableString, as changing the string part of
//a pickablestring implies that it's picking box needs to be recalculated.
template<class PB=null_pick>
class PickableString : public Drawable, public Pickable, public PB {
	string str;
	TextFormat format;
public:
	Drawable_Rects pick_rects;

	PickableString(string s,nehe::TextFormat format, PB pb=PB()) 
		: str(s),format(format), pick_rects(text_to_rect(format,s.c_str())), PB(pb) {}
	void p_draw() { pick_rects.draw(); }
	void pick() { PB::pick(); }
	void draw() { nehe::glPrintf(format, str.c_str());}
	void set_string(string s) { str=s; pick_rects=text_to_rect(format,s.c_str()); }
	void set_format(nehe::TextFormat f) { format=f; pick_rects=text_to_rect(format,s.c_str()); }
	string get_string() const { return str; }
	TextFormat get_format() const { return format; }
	float height() const { return format.ySpace*boost_ex::split(str,'\n').size(); }
};


//abstract because pick(int i) is undefined
class String_Vector : public Multi_Picker, public Drawable, public vector<PickableString<> > {
protected:
	GLfloat ySpace;
public:

	String_Vector(vector<string> strings, nehe::TextFormat format, GLfloat ySpace=.16) :
	ySpace(ySpace) {		
		for(vector<string>::const_iterator i=strings.begin();i!=strings.end();i++) {
			push_back(PickableString<>(*i,format));
		}
	}

	virtual int size() { return vector<PickableString<null_pick> >::size(); }
	
	virtual void p_draw(int base) {		
		glPushMatrix();
		for(iterator i=begin();i!=end();i++) {
			glLoadName(base);
			base++;
			i->p_draw();
			glTranslatef(0,-i->height(),0);
		}
		glPopMatrix();
	}

	virtual void draw() {	
		glPushMatrix();
		for(iterator i=begin();i!=end();i++) {
			i->draw();
			glTranslatef(0,-i->height(),0);
		}
		glPopMatrix();

	}

};

struct VanishingList : public String_Vector {
	VanishingList(vector<string> strings, nehe::TextFormat format, GLfloat ySpace=.16) :
		String_Vector(strings,format,ySpace) {}

	void pick(int i) {
		using namespace win_tools;

		if(g_mouse.lClick()) erase(begin()+i);
	}
};


struct Bound_String_Vector :  public String_Vector, public Float_Rect {
protected:
	int shift;
public:
	void inc_shift() {
		if(shift+begin()!=end()) shift++;
	}
	void dec_shift() {
		if(shift) shift--;
	}

	iterator begin_vis() {
		return begin()+shift;
	}

	struct strings_out_of_bounds {};
protected:

	bool right_out(const Float_Rect & test, const Float_Rect & container) {
		return test.tx<container.tx;
	}

	//zero if a<b.
	float extra(const Float_Rect & test, const Float_Rect & container) {
		float t = test.bx-container.bx;
		return t > 0 ? t : 0;
	}

	int make_in_bounds_y() {
		int rval=0;
		GLfloat total=0;

		for(const_iterator i = begin_vis();i!=end();i++) {
			GLfloat temp = total+i->height();
			int dif=ty-i->get_format().y;
			int h=height();
			if(temp > height()-dif) {
				return rval;
			}
			total=temp;
			rval++;
		}
		return rval;
	}

	iterator end_vis() {
		return begin()+shift+make_in_bounds_y();
	}

	void make_in_bounds_x(PickableString<> &s) {
		using namespace boost_ex;
		Float_Rect me=*static_cast<Float_Rect *>(this);

		string str = s.get_string();

		vector<string> strings = split(str,'\n');
		
		bool altered=false;

		for(int i=0;i<strings.size();i++) {
			
			Float_Rect him = static_cast<Float_Rect>(s.pick_rects[i]);
			if(right_out(him,me)) throw strings_out_of_bounds();
			
			GLfloat extra_len=extra(him,me);
			if(extra_len) {
				altered=true;

				GLfloat intra=him.length()-extra_len;
				
				int l=strings[i].length();
				int newlength=(intra/him.length())*l;
		
				strings[i].resize(newlength);
				if(newlength>5) {
					*(strings[i].end()-1)='.';
					*(strings[i].end()-2)='.';
					*(strings[i].end()-3)='.';
				}
			}
		}
		if(altered) s.set_string(join(strings,"\n"));
	}

public:


	Bound_String_Vector(Float_Rect f, vector<string> strings, TextFormat format, 
		GLfloat ySpace=.16) : Float_Rect(f), String_Vector(strings,format,ySpace),
		shift(0) {}

	Bound_String_Vector(Float_Rect f, vector<string> strings, float x, float y) :
		String_Vector(strings,TextFormat(f.tx+x,f.ty-y),.16f), 
		Float_Rect(f),shift(0)  {}

	virtual void p_draw(int base) {		
		glPushMatrix();
		base+=shift;
		for(iterator i=begin_vis();i!=end_vis();i++) {
			glLoadName(base);
			base++;
			i->p_draw();
			glTranslatef(0,-i->height(),0);
		}
		glPopMatrix();
	}

	virtual void draw() {
		iterator i;
		int top=make_in_bounds_y();

		for(i=begin_vis();i!=end_vis();i++) {
			make_in_bounds_x(*i);
		}

		glPushMatrix();
		for(i=begin_vis();i!=end_vis();i++) {
			i->draw();
			glTranslatef(0,-i->height(),0);
		}
		glPopMatrix();

	}
};


struct VanishingList2 : public Bound_String_Vector {
	VanishingList2(Float_Rect f, vector<string> strings, TextFormat format, GLfloat ySpace=.16) :
		Bound_String_Vector(f,strings,format,ySpace) {}

	VanishingList2(Float_Rect f, vector<string> strings, float x=.2f, float y=.2f,
		rgba color=rgba()) :
	Bound_String_Vector(f,strings,TextFormat(f.tx+x,f.ty-y, color),.16f) {}

	void pick(int i) {
		using namespace win_tools;
		if(g_mouse.lClick()) erase(begin()+i);
	}
};


}
#endif