//So here are some little definition that I threw together 
//to make playing with the boost regex library easier.

#ifndef SIMPLE_REGEXP_H
#define SIMPLE_REGEXP_H

#include <string>
#include <sstream>
#include <boost/regex.hpp> //Mmm... regular expressions...

namespace std_ex {

using namespace std;

//////////////////////////////////////////////////////
//Some common conversion functions:
//Often used to turn the results of a regex search into nummerical data 

inline int make_int(string s) {
	istringstream is(s);
	int rval;
	is >> rval;
	return rval;
}
inline float make_int(const string::const_iterator & start, 
						const string::const_iterator & end) {
	return make_int(string(start,end));
}

inline float make_float(string s) {
	istringstream is(s);
	float rval;
	is >> rval;
	return rval;
}
inline float make_float(const string::const_iterator & start, 
						const string::const_iterator & end) {
	return make_float(string(start,end));
}



///////////////////////////////////////////////////
//istream_to_string: convert an istream to a string.
//
//Taken (with minor modification) from the boost 1.29 regex examples.
//Handy to have around because you can't do regex matching on a istream,
//but you can match a string.
inline void istream_to_string(istream& is, string &s)
{
   s.erase();
   s.reserve(is.rdbuf()->in_avail());
   char c;
   while(is.get(c))
   {
      if(s.capacity() == s.size())
         s.reserve(s.capacity() * 3);
      s.append(1, c);
   }
}

}

namespace boost_ex {

using namespace std_ex;
using namespace boost;

//The 'what' result that stores the matched portioned of a string
//is commonly of type string_results.
typedef match_results<string::const_iterator> string_results;


/////////////////////////////////////////////////////
//simple_search is a wrapper for the regex matching function
//
//quick example:  simple_search("bob was here",regex("(.*) was"),1) 
//				   returns "bob"
//				  given i=0 instead of 1, it would have returned "bob was".

//Thrown if the simple search fails
struct simple_search_error : public string { simple_search_error(string s) : string(s) {} };

inline string simple_search(string s, regex e,int i=0) {
	string_results what;
	if(!regex_search((string::const_iterator)s.begin(),(string::const_iterator)s.end(),what,e))
	{
		throw simple_search_error("simple string match failed with string \"" + s + "\"");
	}
	return string(what[i].first,what[i].second);
}


inline bool is_float(const string s) {
	string_results what;
	return regex_search((string::const_iterator)s.begin(),(string::const_iterator)s.end(),what,
		regex("^-?\\d*(\\.\\d+)?$")) &&
		string(what[0].first,what[0].second)!="";
}

inline bool is_int(const string s) {
	string_results what;
	return 
		regex_search((string::const_iterator)s.begin(),(string::const_iterator)s.end(),what,
		regex("^-?\\d*$"))			&&
		string(what[0].first,what[0].second)!="";
}

inline vector<string> split(const string &s,char mark) {
	vector<string> rval;
	string_results what;
	string::const_iterator start=s.begin();
	string::const_iterator end=s.end();
	
	while(start!=end) {
		if(regex_search(start,end,what,regex(string("")+mark)) ) {
			rval.push_back(string(start,what[0].first));
			start=what[0].second;
		}
		else {
			rval.push_back(string(start,end));
			break;
		}

	}
	return rval;
}

inline string join(const vector<string> & v, string mark) {
	string rval;
	if(v.size()){
		for(vector<string>::const_iterator i=v.begin();i!=v.end()-1;i++) {
			rval+=*i+mark;
		}
		rval+=*i;
	}

	return rval;
}

}

#endif