#ifndef STAM_H
#define STAM_H

#include <set>
#include <sven.h>

namespace Stam {

const float smallf = 0.0000001f;


//simulation data management
void free_data ( void );
void clear_data ( void );
void init_data();

//UI entry points
void key_func ( unsigned char key, int x, int y );
void mouse_func ( int button, bool down, int x, int y );
void motion_func ( int x, int y );
void set_bound( int mx, int my);
void kill_bound( int mx, int my);


//general gl housekeeping
//(this guy needs to be called before we start trying to display
void reshape_func ( int width, int height );

//as usual "idle" means "update the simulation"
void idle_func ( void );
void display_func ( void );
void draw_bounds(); //kept seperate from display so i can draw the bounds over the npr rendering
void draw_density(); //npr needs to see this so it can use it to pick brush colors

//various constants of the simulation
extern float epsilon;
extern float dt, diff, visc;
extern float ddecay, dalpha;
extern int selected_dye;

extern sven::rgb bound_color;

//these define the boundries of the sim
//(square edges are also assummed)
extern bool * blocked;

}

namespace display {
extern bool show_bounds;
extern bool show_velocity;
extern bool show_density;
extern bool show_points;
extern bool show_brushes;
extern bool show_triangles;

}

#include "wx/glcanvas.h"

struct UI_Abstract {
	virtual void OnMouse(wxMouseEvent& event) =0;
	virtual void OnChar(wxKeyEvent& event) =0;
	virtual ~UI_Abstract() {}
};


struct UI_Bounds : public UI_Abstract {

	void OnMouse(wxMouseEvent& event) {
		if(event.m_leftDown) Stam::set_bound(event.GetX(),event.GetY());
		if(event.m_rightDown) Stam::kill_bound(event.GetX(),event.GetY());
		Stam::motion_func(event.GetX(),event.GetY());
	}

	void OnChar(wxKeyEvent& event) {
		Stam::key_func(
			(unsigned char) event.m_keyCode,
			event.GetX(),
			event.GetY()
			);
	}

};

extern UI_Abstract * currentUI;
extern UI_Bounds boundsUI;




#endif