//Stam's own implmentation of his solver, now hacked around a bit
//by me (Sven).  Note that this code uses sparse linear solvers, rather
//than FFTs.
//All the comments are mine, for good documentation, read Stam's paper.

#include <stdlib.h>

#include <win_globals.h>
#include <sstream>
#include <math.h>
#include <sven.h>

#include "stam.h"

namespace Stam {

bool * blocked;

//Macro heavy old-school C code ...
#define IX(i,j) ((i)+(N+2)*(j))
#define SWAP(foo0,foo) {float * tmp=foo0;foo0=foo;foo=tmp;}
#define FOR_EACH_CELL for ( i=1 ; i<=N ; i++ ) { for ( j=1 ; j<=N ; j++ ) { 
#define END_FOR }}


//I suppose it might be slightly more efficient to use function objects,
//but for whatever reason right now I prefer to just use
//function pointers.
typedef void (advect_func)( int N, int b, float * d, float * d0, float * u, float * v, float dt );

void add_source ( int N, float * x, float * s, float dt )
{
	int i, size=(N+2)*(N+2);
	for ( i=0 ; i<size ; i++ ) x[i] += dt*s[i];
}


//So I've pretty agressivly simplified this method,
//all I want it to do is take care of the edge boundry conditions
//
void set_bnd ( int N, int b, float * x )
{
	int i,j;
	for ( i=1 ; i<=N ; i++ ) {
		x[IX(0  ,i)] = b==1 ? 0 : x[IX(1,i)];
		x[IX(N+1,i)] = b==1 ? 0 : x[IX(N,i)];
		x[IX(i,0  )] = b==2 ? 0 : x[IX(i,1)];
		x[IX(i,N+1)] = b==2 ? 0 : x[IX(i,N)];
	}

	FOR_EACH_CELL
		if(!blocked[IX(i,j)]) continue;
		if( (b==1) || (b==2) ) {
			x[IX(i,j)]=0;
		}
		else {
			int count  =0;
			float total=0;
			if(!blocked[IX(i+1,j)]) { count++; total+=x[IX(i+1,j)]; }
			if(!blocked[IX(i-1,j)]) { count++; total+=x[IX(i-1,j)]; }		
			if(!blocked[IX(i,j+1)]) { count++; total+=x[IX(i,j+1)]; }
			if(!blocked[IX(i,j-1)]) { count++; total+=x[IX(i,j-1)]; }
			if(count) total/=count;
			x[IX(i,j)]=total;
		}
	END_FOR

}

//Stam's origional solver had an asymetry which caused bugs given
//my boundry implmentation.  
void lin_solve ( int N, int b, float * x, float * x0, float a, float c )
{
	int i, j, k;

	int size = (N+2)*(N+2);
	float * temp	= (float *) malloc ( size*sizeof(float) );

	for ( k=0 ; k<20 ; k++ ) {
		FOR_EACH_CELL
			temp[IX(i,j)] = (x0[IX(i,j)] + a*(x[IX(i-1,j)]+x[IX(i+1,j)]+x[IX(i,j-1)]+x[IX(i,j+1)]))/c;
		END_FOR

		FOR_EACH_CELL
			x[IX(i,j)] = temp[IX(i,j)];
		END_FOR

		set_bnd ( N, b, x );
	}

	free(temp);
}

void diffuse ( int N, int b, float * x, float * x0, float diff, float dt )
{
	float a=dt*diff*N*N;
	lin_solve ( N, b, x, x0, a, 1+4*a );
}

void linear_advect ( int N, int b, float * d, float * d0, float * u, float * v, float dt ) {
	int i, j, i0, j0, i1, j1;
	float x, y, s0, t0, s1, t1, dt0, vx, vy, tleft,t,tnext,distx,disty;

	dt0 = dt*N;


	FOR_EACH_CELL
	
		if(blocked[IX(i,j)]) continue;

		tleft=dt0;
		x=i;y=j;		

		while(tleft>smallf) {

			//enforce boundry contraints
			if (x<0.5f) x=0.5f; if (x>N+0.5f) x=N+0.5f; 
			if (y<0.5f) y=0.5f; if (y>N+0.5f) y=N+0.5f; 


			i0=(int)x; i1=i0+1;
			j0=(int)y; j1=j0+1;
			s1 = x-i0; s0 = 1-s1; t1 = y-j0; t0 = 1-t1;

			vx = -(s0*(t0*u[IX(i0,j0)]+t1*u[IX(i0,j1)])+
				   s1*(t0*u[IX(i1,j0)]+t1*u[IX(i1,j1)]));
					
			vy = -(s0*(t0*v[IX(i0,j0)]+t1*v[IX(i0,j1)])+
				   s1*(t0*v[IX(i1,j0)]+t1*v[IX(i1,j1)]));


			float speed2=vx*vx+vy*vy; 
			if(speed2>smallf) tnext=.5/sqrt(speed2);
			else tnext=tleft;

			t=tnext > tleft ? tleft : tnext;
			tleft-=t;


			x+=t*vx;
			y+=t*vy;
		}


		if (x<0.5f) x=0.5f; if (x>N+0.5f) x=N+0.5f; 
		if (y<0.5f) y=0.5f; if (y>N+0.5f) y=N+0.5f; 


		i0=(int)x; i1=i0+1;
		j0=(int)y; j1=j0+1;
		s1 = x-i0; s0 = 1-s1; t1 = y-j0; t0 = 1-t1;

		d[IX(i,j)] = s0*(t0*d0[IX(i0,j0)]+t1*d0[IX(i0,j1)])+
					 s1*(t0*d0[IX(i1,j0)]+t1*d0[IX(i1,j1)]);
	END_FOR
	set_bnd ( N, b, d );
}


//The eventual idea is to have a version of advect that uses the monotonic cubic iterpolation
//described in "Visual Simulation of Smoke"
advect_func * advect=&linear_advect;

void project ( int N, float * u, float * v, float * p, float * div )
{
	int i, j;

	FOR_EACH_CELL
		div[IX(i,j)] = -0.5f*(u[IX(i+1,j)]-u[IX(i-1,j)]+v[IX(i,j+1)]-v[IX(i,j-1)])/N;
		p[IX(i,j)] = 0;
	END_FOR	
	set_bnd ( N, 0, div ); set_bnd ( N, 0, p );

	lin_solve ( N, 0, p, div, 1, 4 );

	FOR_EACH_CELL
		u[IX(i,j)] -= 0.5f*N*(p[IX(i+1,j)]-p[IX(i-1,j)]);
		v[IX(i,j)] -= 0.5f*N*(p[IX(i,j+1)]-p[IX(i,j-1)]);
	END_FOR
	set_bnd ( N, 1, u ); set_bnd ( N, 2, v );
}

float epsilon=2;

void add_vorocity(int N, float * u, float * v, float dt) {

//vort is just the curl of the velocity field
#define vort(i,j) (( -0.5*( v[IX(i+1,j)]-v[IX(i-1,j)] + u[IX(i,j-1)]-u[IX(i,j+1)] )*N ))
	
	int i, j;

	int size = (N+2)*(N+2);

	float *fx		= (float *) malloc ( size*sizeof(float) );
	float *fy		= (float *) malloc ( size*sizeof(float) );

	FOR_EACH_CELL
		if(i==1 || j==1 || i==N || j==N) continue;

		float dvx = 0.5*(vort(i+1,j)-vort(i-1,j));
		float dvy = 0.5*(vort(i,j+1)-vort(i,j-1));

		float mag=sqrt(dvx*dvx+dvy*dvy);

		//apply a force equal to epsilon * cross(N,vor)
		fx[IX(i,j)]=mag > smallf ? epsilon*dt*dvy*vort(i,j)/(mag*N) : 0;
		fy[IX(i,j)]=mag > smallf ? epsilon*dt*dvx*vort(i,j)/(mag*N) : 0;
	END_FOR

	set_bnd ( N, 0, fx ); set_bnd ( N, 0, fy );


	FOR_EACH_CELL
		if(i==1 || j==1 || i==N || j==N) continue;

		if(u[IX(i,j)]) u[IX(i,j)]+=fx[IX(i,j)];
		if(v[IX(i,j)]) v[IX(i,j)]-=fy[IX(i,j)];
	END_FOR

	set_bnd ( N, 1, u ); set_bnd ( N, 2, v );

free(fx);
free(fy);

#undef vort

}

void dens_step ( int N, float * x, float * x0, float * u, float * v, float diff, float dt )
{
	add_source ( N, x, x0, dt );
	SWAP ( x0, x ); diffuse ( N, 0, x, x0, diff, dt );
	SWAP ( x0, x ); advect ( N, 0, x, x0, u, v, dt );

}

void vel_step ( int N, float * u, float * v, float * u0, float * v0, float visc, float dt )
{
	add_source ( N, u, u0, dt ); add_source ( N, v, v0, dt );
	SWAP ( u0, u ); diffuse ( N, 1, u, u0, visc, dt );
	SWAP ( v0, v ); diffuse ( N, 2, v, v0, visc, dt );
	project ( N, u, v, u0, v0 );
	SWAP ( u0, u ); SWAP ( v0, v );
	advect ( N, 2, v, v0, u0, v0, dt ); advect ( N, 1, u, u0, u0, v0, dt ); 
	project ( N, u, v, u0, v0 );
}

}
